""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license are defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""

import random, math

class OMmanager():
    """Represents all O & M activities for one farm. Called each game tick"""
    def __init__(self,strategy,farm,game):
        self.strategy=strategy
        self.farm=farm
        self.g=game
        self.repairLog=[]
        self.failureLog=[]
        self.previousRepair=0
        self.profit_per_hour_per_turbine = None
        self.totalSpent = self.g.costs.costlist['ConditionBasedMaintenance_upfront'] if self.strategy=='Condition Based' else 0
        self.totalDownTime = 0
        self.reportNumberFailures = {'farm':0,'structure':0,'turbine':0}
        self.reportDownTime = {'farm':0,'structure':0,'turbine':0}
        self.failuresPerTurbinePerYear = 0
        self.annualOandMcost = 0

    def setup(self,farm):
        self.profit_per_hour_per_turbine = int(math.ceil((self.farm.myTurbineList[0].readPowerCurve(self.farm.myCell.meanU90) * 1) * farm.sellingPrice['First10Years']))  # energy produced by one turbine in 1 hour with average wind speed

    def maintain(self): #for new failures
        if self.strategy=='Corrective':
            self.corrective()
        if self.strategy=='Condition Based':
            self.conditionBased()
        if self.strategy=='Scheduled_Monthly':
            self.scheduledMonthly()
        if self.g.windClock % 525600 == 0:
            self.annualInspection()
        self.failuresPerTurbinePerYear = round((len(self.failureLog)/len(self.farm.myTurbineList))/math.ceil(self.farm.status['age']/525600),2)
        self.annualOandMcost = round(self.farm.costs['OPEX']/(self.farm.status['age']/525600),2)

    def vesselCost(self,FailType):
        try: #For turbine failures
            duration = self.g.costs.costlist['duration_' + str(FailType)] #Duration of each repair
        except:
            duration = 2 #for all other failures - come back to this
        vessel_cost = 0
        vessel_type = None
        if 'Minor_Repair' in FailType:
            vessel_type = 'CTV'
        elif 'Major_Repair' in FailType:
            vessel_type = 'FSV'
        elif 'Major_Replacement' in FailType:
            vessel_type = 'HLV'
        elif 'farm_grid' in FailType:
            vessel_type = 'AHTS'
        else: #For non-turbine failures
            vessel_type = 'FSV'
        vessel_cost = duration * self.g.costs.costlist['vessel_'+str(vessel_type)]
        return duration, vessel_cost, vessel_type

    def annualInspection(self):
        vessel_cost = self.g.costs.costlist['duration_Annual_Inspection_perturbine'] * self.g.costs.costlist['vessel_CTV']
        cost=len(self.farm.myTurbineList) *self.g.costs.costlist['Annual_Inspection_perturbine'] + vessel_cost
        print('Annual maintenance on {}. (Cost: {:,})'.format(self.farm.name, int(cost)))
        self.g.wallet.withdraw(cost,str(self.g.tool.convertTime(self.g.windClock)['Short Date'])+' : Annual Maintenance on '+str(self.farm.name))
        self.farm.costs['OPEX'] += cost

    def monthlyRepairs(self):
        if self.g.windClock % (525600 / 12) == 0 and self.g.windClock != 0: #Is it that time of the month again?
            cost =0
            max_vessel_cost_required = (0,) #vessel cost will be the largest most expensive required vessel for this monthly visit
            duration_all = 0
            repairMonthly={}#save as another dictionary, otherwise loop won't run as the size of the dict changes.
            for Loc, FailType in self.farm.failures.items():
                repairMonthly[Loc]=FailType
            i=0 #For logging purposes, indicate which repair is being done first, as to only include the cost of the vessel once! - This concept is not exactly correct. needs to change in future version.
            for Loc, FailType in repairMonthly.items():
                self.g.repairVessel.process(self.farm, Loc, FailType)
                repairCost = self.g.costs.costlist[FailType]
                cost += repairCost  #Increase cost for repair of each failure
                self.g.windFarmGrid.failureList.remove([self.farm, Loc, FailType])
                if i==0:
                    self.logRepairs(Loc, FailType, cost, 'full repair')
                    i=1 #Logged the cost of the vessel for this batch
                else:
                    self.logRepairs(Loc, FailType,repairCost, 'full repair')
                duration, vessel_cost, vessel_type = self.vesselCost(FailType)
                duration_all += duration
                max_vessel_cost_required += (self.g.costs.costlist['vessel_'+str(vessel_type)],)
            vessel_cost = max(max_vessel_cost_required) * duration_all*0.75 #Take largest vessel required, send it out there for the duration of all repairs, assumed rate to account for efficiency to be updated in future
            cost +=vessel_cost
            self.farm.costs['OPEX']+=vessel_cost
            self.g.wallet.withdraw(cost, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Repair of ' + self.farm.name + ' Monthly repairs ')
            self.g.gui.updateCellColor(self.farm.myCell)
            #print('\nMonthly repairs completed for {} : year {} month {}: ''{} Repairs for cost of {:,}'.format(self.farm.name, self.g.tool.convertTime(self.g.windClock)['years'],self.g.tool.convertTime(self.g.windClock)['months'],len(repairMonthly),int(cost)))

    def corrective(self):
        if self.farm.failures!={}: #will run if there are any failures, but nothing really happens unless the failure is new.
            repairnow={}
            for Loc,FailType in self.farm.failures.items():
                if Loc.notification['fail']=='new':
                    Loc.notification['fail'] = 'old' #If user decides to repair later, this notification doesn't appear anymore.
                    self.logFailures(Loc,FailType)
                    self.g.gui.updateCellColor(self.farm.myCell)
                    if self.farm.waveHeight<=2.0 and self.farm.windSpeed <= 15 and 'Major_Replacement' in FailType:
                        self.g.pause()
                        print("\n<!> {} {} has had a {} failure. Weather is good - Repair now? Y/N <!>".format(self.farm.name, Loc.name, FailType))
                        duration, vessel_cost, vessel_type = self.vesselCost(FailType)
                        repair=input('\t\tCost Now:      €{:10,.0f}\n\t\tCost Later:    €{:10,.0f} (while doing other scheduled repairs)\n\t\tProfit Losses: €{:10,.0f} (per day for every day without repair! Approximately.)\n'.format(self.roundTo(vessel_cost+self.g.costs.costlist[FailType],1000),self.roundTo(self.g.costs.costlist[FailType],1000),self.roundTo(self.profit_per_hour_per_turbine*24,1000)))
                        while repair!='N' and repair!='n' and repair!='Y' and repair!='y':
                          repair = input('Would you like to repair this failure immediately? Y/N')

                        if repair.lower() =='y':
                            repairnow[Loc]=FailType
                            print('Repair of {} {} complete!'.format(Loc.name, FailType))
                        else:
                            print('Go into main menu [f] at any time to repair this failure. Otherwise it will be repaired at the end of the month.')
                            self.g.windFarmGrid.listFailures(self.farm, Loc, FailType) #save for later
                        self.g.unpause()
                    elif self.farm.waveHeight<=2.0 and self.farm.windSpeed <= 15 and 'Major_Replacement' not in FailType and FailType != 'structure_localscour': #Weather is good - do non-major replacement repair now
                        repairnow[Loc] = FailType
                    elif FailType == 'structure_localscour':
                        self.g.gui.updateCellColor(self.farm.myCell)
                        repairnow[Loc] = FailType
                    else: #If weather is not good, save for later
                        self.g.windFarmGrid.listFailures(self.farm, Loc, FailType)  # save for later
            for Loc, FailType in repairnow.items():  #Repair immediately if weather is good
                if FailType != 'structure_localscour':
                    self.g.repairVessel.process(self.farm, Loc, FailType)
                    duration, vessel_cost, vessel_type = self.vesselCost(FailType)
                    cost = vessel_cost + self.g.costs.costlist[FailType]
                    self.g.wallet.withdraw(cost, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Repair of ' + self.farm.name + ' ' + FailType)
                    self.g.windFarmGrid.failureHistory.append([self.farm, Loc, FailType])  # failure goes straight intohistory
                    self.logRepairs(Loc, FailType, cost, 'full repair')
                    self.g.gui.updateCellColor(self.farm.myCell)
                elif FailType == 'structure_localscour':
                    self.scourRepair(Loc, FailType)
                    self.g.gui.updateCellColor(self.farm.myCell)
        self.monthlyRepairs() #At end of month, repair everyting

    def scheduledMonthly(self):
        if self.farm.failures!={}: #will run if there are any failures. The user does not have the option to repair the failure at the moment it occurs.
            repairnow = {}
            for Loc, FailType in self.farm.failures.items():
                if Loc.notification['fail']=='new':
                    Loc.notification['fail'] = 'old'
                    self.logFailures(Loc, FailType)
                    self.g.gui.updateCellColor(self.farm.myCell)
                    if FailType == 'structure_localscour':
                        repairnow[Loc] = FailType
                        self.g.gui.updateCellColor(self.farm.myCell)
                    else:
                        self.g.windFarmGrid.listFailures(self.farm, Loc, FailType)
            #Scour repairs are done immediately
            for Loc, FailType in repairnow.items():  # Repair immediately if weather is good
                self.scourRepair(Loc, FailType)
                self.g.gui.updateCellColor(self.farm.myCell)
        self.monthlyRepairs()

    def conditionBased(self):
        if self.farm.failures != {}: #will run if there are any failures, but nothing really happens unless the ailure is new.
            repairnow = {}
            prevent={}
            for Loc, FailType in self.farm.failures.items():
                if Loc.notification['fail'] == 'new' and random.random() <= self.farm.myFailureRates.otherfailures['conditionbasedmonitoring_efficiency'] and FailType in self.farm.myFailureRates.preventableFailures: #If there is a new failure, set a random chance that this failure is actually impending (give the user a chance to catch the failure before it happens, because of the investment they made into monitoring.)
                    Loc.notification['fail'] = 'old'
                    self.logFailures(Loc, FailType)
                    if self.farm.waveHeight <= 2.0 and self.farm.windSpeed <= 15 and 'Major_Replacement' in FailType:
                        self.g.pause()
                        print('<!> {} {} - Monitoring system has reported possible need for {}. Weather is good - Repair now? Y/N <!>'.format(self.farm.name, Loc.name, FailType))
                        duration, vessel_cost, vessel_type = self.vesselCost(FailType)
                        repair = input('\t\tCost Now:      €{:10,.0f}\n\t\tCost Later:    €{:10,.0f} (while doing other scheduled repairs)\n\t\tProfit Losses: €{:10,.0f} (per day for every day without repair! Approximately.)\n'.format(self.roundTo(vessel_cost+self.g.costs.costlist[FailType] * self.g.costs.costlist['RATE_preventative_repair'],1000),self.roundTo(self.g.costs.costlist[FailType],1000),self.roundTo(self.profit_per_hour_per_turbine*24,1000)))
                        while repair != 'Y' and repair != 'y' and repair != 'n' and repair != 'N':
                            repair = input('Would you like to prevent this failure by going out to do maintenance work? Y/N')
                        if repair=='Y' or repair=='y':
                            prevent[Loc]=FailType
                            print('Preventative repair of {} {} complete!'.format(Loc.name, FailType))
                        else:
                            print('Go into main menu [f] at any time to repair this failure. Otherwise it will be repaired at the end of the month.')
                            self.g.windFarmGrid.listFailures(self.farm, Loc, FailType)  # save for later
                        self.g.unpause()
                    elif self.farm.waveHeight<=2.0 and self.farm.windSpeed <= 15 and 'Major_Replacement' not in FailType: #Weather is good - do non-major replacement preventative repair now
                        prevent[Loc] = FailType
                    else: #If weather is not good, save for later
                        self.g.windFarmGrid.listFailures(self.farm, Loc, FailType)  # save for later
                elif Loc.notification['fail'] == 'new': #and random.random() > self.farm.myFailureRates.otherfailures['conditionbasedmonitoring_efficiency']:
                    Loc.notification['fail'] = 'old'
                    self.logFailures(Loc, FailType)
                    self.g.gui.updateCellColor(self.farm.myCell)
                    if self.farm.waveHeight <= 2.0 and self.farm.windSpeed <= 15 and 'Major_Replacement' in FailType:
                        self.g.pause()
                        print("<!> {} {} has had a {} failure. Weather is good - Repair now? Y/N <!>".format(self.farm.name, Loc.name, FailType))
                        duration, vessel_cost, vessel_type = self.vesselCost(FailType)

                        repair = input('\t\tCost Now:      €{:10,.0f}\n\t\tCost Later:    €{:10,.0f} (while doing other scheduled repairs)\n\t\tProfit Losses: €{:10,.0f} (per day for every day without repair! Approximately.)\n'.format(self.roundTo(vessel_cost+self.g.costs.costlist[FailType],1000),self.roundTo(self.g.costs.costlist[FailType],1000),self.roundTo(self.profit_per_hour_per_turbine*24,1000)))
                        while repair != 'N' and repair != 'n' and repair != 'Y' and repair != 'y':
                            repair = input('Would you like to repair this failure immediately? Y/N')
                        if repair == 'Y' or repair == 'y':
                            repairnow[Loc] = FailType
                            print('{} {} repaired immediately!\n'.format(Loc.name, FailType))
                        else:
                            print('Go into main menu [f] at any time to repair this failure. Otherwise it will be repaired at the end of the month.')
                            self.g.windFarmGrid.listFailures(self.farm, Loc, FailType)           # save for later
                        self.g.unpause()
                    elif self.farm.waveHeight<=2.0 and self.farm.windSpeed <= 15 and 'Major_Replacement' not in FailType and FailType != 'structure_localscour': #Weather is good - do non-major replacement preventative repair now
                        repairnow[Loc] = FailType
                    elif FailType == 'structure_localscour':
                        repairnow[Loc] = FailType
                        self.g.gui.updateCellColor(self.farm.myCell)
                    else:#If weather is not good, save for later
                        self.g.windFarmGrid.listFailures(self.farm, Loc, FailType)               # save for later

            for Loc, FailType in repairnow.items():                                                 #Repair if the user decided to repair immediately.
                if FailType != 'structure_localscour':
                    self.g.repairVessel.process(self.farm, Loc, FailType)
                    duration, vessel_cost, vessel_type = self.vesselCost(FailType)
                    cost = vessel_cost + self.g.costs.costlist[FailType]
                    self.g.wallet.withdraw(cost, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Repair of ' + self.farm.name + ' ' + FailType)
                    self.g.windFarmGrid.failureHistory.append([self.farm, Loc, FailType])            # failure goes straight intohistory
                    self.logRepairs(Loc, FailType, cost, 'full repair')
                    self.g.gui.updateCellColor(self.farm.myCell)
                elif FailType == 'structure_localscour':
                    self.scourRepair(Loc, FailType)
                    self.g.gui.updateCellColor(self.farm.myCell)
            for Loc, FailType in prevent.items():                                                   #Do repair work on preventative repairs is the user decides to do so immediately
                self.g.repairVessel.process(self.farm, Loc, FailType)
                duration, vessel_cost, vessel_type = self.vesselCost(FailType)
                cost = vessel_cost + self.g.costs.costlist['RATE_preventative_repair'] * self.g.costs.costlist[FailType]
                self.g.wallet.withdraw(cost, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Preventative Repair of ' + self.farm.name + ' ' + FailType)
                self.g.windFarmGrid.failureHistory.append([self.farm, Loc, FailType])  # failure goes straight intohistory
                self.logRepairs(Loc, FailType, cost, 'preventative')
        self.monthlyRepairs()  # At end of month, repair everyting

    def manageFailures(self,toRepair,index,repair_session_time): #Which failure do we want to manage? Do the repair
        Farm, Loc, FailType = toRepair
        if self.farm.waveHeight <=2.0 and self.farm.windSpeed <= 15:#check weather at farm. if waves are too high, can't repair!
            if FailType=='construction_accident': #If failure is during construction, repair regardless of O&M strategy
                self.g.repairVessel.process(Farm, Loc, FailType)
                duration, vessel_cost, vessel_type = self.vesselCost(FailType)
                cost= vessel_cost + self.g.costs.costlist['construction_accident']
                self.g.wallet.withdraw(cost, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Repair of ' + FailType)
                del self.g.windFarmGrid.failureList[index]
                self.logRepairs(Loc, FailType, cost, 'full repair')
            elif self.strategy=='Corrective' or self.strategy == 'Condition Based':
                self.g.repairVessel.process(Farm, Loc, FailType)
                if repair_session_time > self.previousRepair: #If this is a new repair session, need to charter the repair vessel
                    duration, vessel_cost, vessel_type = self.vesselCost(FailType) #This does not use the most expensive vessel, ought to be changed!
                elif repair_session_time == self.previousRepair: #If we are still in the same repair session (game pauses when we go into a session), then keep the previous vessel chartered
                    vessel_cost = 0
                cost = vessel_cost + self.g.costs.costlist[FailType]
                print('Repair of {} {} complete! cost: {:,}'.format(Loc.name, FailType,self.roundTo(cost,1000)))
                self.g.wallet.withdraw(cost, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Repair of ' + FailType)
                del self.g.windFarmGrid.failureList[index]
                self.previousRepair=repair_session_time
                self.logRepairs(Loc, FailType, cost, 'full repair')
            elif self.strategy=='Scheduled_Monthly':
                print('Must wait until scheduled monthly visit to repair {} {}'.format(Loc.name, FailType))
            self.g.gui.updateCellColor(self.farm.myCell)
        else:
            print('\nThe wave height is too high ({} m), try again later.'.format(self.farm.waveHeight))

    def logFailures(self,Loc,FailType): #Keep a record of all failures that occured, and their impact on the game
        t = self.g.windClock
        if str(Loc.__class__) == "<class 'mTurbine.Turbine'>":  # if the failure occured at a turbine
            N=1 #number of turbines affected
            self.reportNumberFailures['turbine'] += 1
        elif str(Loc.__class__) == "<class 'mSubstructure.Substructure'>":
            N = 1  # number of turbines affected
            self.reportNumberFailures['structure'] += 1
        else:
            N=len(self.farm.myTurbineList)
            self.reportNumberFailures['farm'] += 1
        failureInfo={'TimeofFailure':t,'TimeofRepair':'Not Repaired yet','Duration':'Not Repaired yet','Year':self.g.tool.convertTime(t)['years'],'Month':self.g.tool.convertTime(t)['months'],'Strategy':self.strategy,'Farm':self.farm.name,'Repair Location':Loc.name,'Repair':FailType,'turbine-hours lost':0,'Profit lost':0,'Number of turbines':N,'Loc':Loc}
        self.failureLog.append(failureInfo)

    def logRepairs(self, loc, repair, cost, repairType): #Keep a record of all repairs done
        t=self.g.windClock
        repairInfo={'Time':t,'Year':self.g.tool.convertTime(t)['years'],'Month':self.g.tool.convertTime(t)['months'],'Farm':self.farm.name,'Strategy':self.strategy,'Repair Location':loc.name,'Repair':repair,'Cost':cost,'Type':repairType} #repair is a dictionary with all repairs completed during one session
        self.totalSpent += cost
        self.farm.costs['OPEX'] += round(cost,0)
        #Calculate values for failure info based on when repair was made
        for failure in self.failureLog: #Search for the failure log for the failure that is being repaired.
            if repairInfo['Repair Location'] == failure['Repair Location']:
                failure['TimeofRepair']=t
                failure['Duration']=t-failure['TimeofFailure'] #Amount of minutes the failure was not repaired. might be zero
                failure['turbine-hours lost'] = (failure['Duration']/60) * failure['Number of turbines']
                failure['Profit lost'] = int(self.profit_per_hour_per_turbine * failure['turbine-hours lost'])
                repairInfo['FailureLog'] = failure

        self.totalDownTime += repairInfo['FailureLog']['Duration']
        self.repairLog.append(repairInfo)  # Add repair to repair log

    def downtime_todate(self): #calculate the amount of downtime in turbine hours to date
        downtimetodate=0
        for failure in self.failureLog:
            if failure['TimeofRepair'] == 'Not Repaired yet':
                downtimetodate += (self.g.windClock - failure['TimeofFailure']) * failure['Number of turbines']
            else:
                downtimetodate += failure['Duration'] * failure['Number of turbines']
        downtimetodate = downtimetodate/60 #Convert minutes to hours
        return downtimetodate #in turbine-hours

    def scourRepair(self, loc, FailType):
        print("\n<!> {} {} has had scour failure as a result of the poor soil and current conditions <!>".format(self.farm.name, loc.name, FailType))
        duration, vessel_cost, vessel_type = self.vesselCost(FailType)
        cost = vessel_cost + self.g.costs.costlist[FailType] * len(self.farm.myTurbineList)
        print("<!> It is predicted that without protection, scour will occur at all of the remaining structures, "
              "so you are required to add scour protection to all structures in the farm at a cost of {:,} per structure. "
              "That is {:,} total with vessel costs. <!>".format(self.roundTo(self.g.costs.costlist[FailType],1000), self.roundTo(cost,1000)))
        self.g.wallet.withdraw(cost, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Repair of all structures at ' + self.farm.name + ' ' + FailType)

        self.g.windFarmGrid.failureHistory.append([self.farm, loc, FailType])  # failure goes straight into history
        self.g.repairVessel.process(self.farm, loc, FailType)
        self.logRepairs(loc, FailType, cost, 'full repair')

        for structure in self.farm.mySubstructureList: #once scour protection is added, it wont occur again
            structure.scourProtection = 'Rock Armouring'
            structure.myStructureFailureRates[FailType] = 0

    def roundTo(self,value,nearest):
        return int(math.ceil(value / nearest)) * nearest

    def endGame(self):
        t = self.g.windClock
        for failure in self.failureLog: #Add duration to unrepaired failures at end of game
            if failure['TimeofRepair'] == 'Not Repaired yet':
                failure['Duration'] = t - failure['TimeofFailure']  # Amount of minutes the failure was not repaired. might be zero

        for failure in self.failureLog:
            loc = failure['Loc']
            # For end of game reporting
            if str(loc.__class__) == "<class 'mTurbine.Turbine'>":  # if the failure occured at a turbine
                self.reportDownTime['turbine'] += failure['Duration']
            elif str(loc.__class__) == "<class 'mSubstructure.Substructure'>":
                self.reportDownTime['structure'] += failure['Duration']
            else:
                self.reportDownTime['farm'] += failure['Duration']


#x=OandMmanager('a')
#x.displayOptions()
