""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license are defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""

import math, mSubstructure, mSubstructureType, mFailureRates

class Optimizer():
    """This class stores methods to provide feedback to the user on how to optimize offshore wind farm design"""

    def __init__(self,game):
        self.g=game #The optimizer needs all information from the game
        self.optimizedFarms = []

    def report(self):
        print('____________________________________________________________________________')
        print('Lets review progress in the past year:')
        self.optimizeOANDM()
        print('____________________________________________________________________________')

    def optimizeOANDM(self): #Every year, the user gets a summary of the incurred O&M costs, and what they would have been using the other two strategies.
        #Go through each farm and calculate alternative strategies for the same failures.
        print('\tIn terms of operation and maintenance, here are some design tips:\n')
        for farm in self.g.windFarmGrid.windFarmList:
            if farm.status['phase'] == 'Operating':
                if farm.OandM.strategy == 'Condition Based':
                    preventedRepairs=0
                    CBM_savings = 0
                    for repair in farm.OandM.repairLog:
                        if repair['Type'] == 'preventative':
                            preventedRepairs+=1
                            CBM_savings += (1 - self.g.costs.costlist['RATE_preventative_repair']) * self.g.costs.costlist[repair['Repair']]  # Find the difference between the cost of the full repair and the preventable repair
                    downtime = farm.OandM.downtime_todate()
                    print('\tAt {}, the total cost of O&M work to date is {:,} (includes initial investment) with a downtime of {:,} turbine-hours.'.format(farm.name, self.roundTo(farm.OandM.totalSpent,10000), self.roundTo(downtime,1000)))
                    print('\t\tUsing Condition Based maintenance allowed for the prevention of {} full blown out failures, and a saving of {:,} in repair costs.'.format(preventedRepairs,self.roundTo(CBM_savings,10000)))
                    estimated_cost_scheduledOandM, estimated_downtime_scheduledOandM=self.monthly_scheduled_estimator(farm,farm.OandM.failureLog)
                    print('\t\tIf monthly scheduled maintenance had been the strategy, the cost would have been {:,} with a downtime of {:,} turbine-hours.\n'.format(self.roundTo(estimated_cost_scheduledOandM,10000),self.roundTo(estimated_downtime_scheduledOandM,1000)))

                if farm.OandM.strategy == 'Corrective':
                    preventablefails=0
                    immediate_repairs=0
                    CBM_savings = 0
                    for repairInfo in farm.OandM.repairLog:#check downtime saved by fixing failures right away as opposed to waiting until the end of the month
                        if repairInfo['FailureLog']['Duration'] == 0:
                            immediate_repairs += 1
                        if repairInfo['Repair'] in farm.myFailureRates.preventableFailures:
                            preventablefails+=1
                            CBM_savings += (1 - self.g.costs.costlist['RATE_preventative_repair']) * self.g.costs.costlist[repairInfo['Repair']]  # Find the difference between the cost of the full repair and the preventable repair
                    downtime = farm.OandM.downtime_todate()
                    print('\tAt {}, the total cost of O&M work to date is {:,} with a downtime of {:,} turbine-hours on {} immediate repairs.'.format(farm.name, self.roundTo(farm.OandM.totalSpent,10000), self.roundTo(downtime,1000),immediate_repairs))
                    print('\t\tIf Condition Based Monitoring had been the strategy, {} failures could have been prevented, possibly saving {:,}.'.format(preventablefails,self.roundTo(CBM_savings,10000)))
                    estimated_cost_scheduledOandM, estimated_downtime_scheduledOandM = self.monthly_scheduled_estimator(farm, farm.OandM.failureLog)
                    print('\t\tIf Monthly Scheduled Maintenance had been the strategy, the cost would have been {:,} with a downtime of {:,} turbine-hours.\n'.format(self.roundTo(estimated_cost_scheduledOandM,10000), self.roundTo(estimated_downtime_scheduledOandM,1000)))

                if farm.OandM.strategy == 'Scheduled_Monthly':
                    months=math.floor(farm.status['age']/43800)
                    preventablefails = 0
                    CBM_savings = 0
                    corrective_cost = 0
                    for repairInfo in farm.OandM.repairLog:  # check downtime saved by fixing failures right away as opposed to waiting until the end of the month
                        if repairInfo['Repair'] in farm.myFailureRates.preventableFailures:
                            preventablefails += 1
                            CBM_savings += (1 - self.g.costs.costlist['RATE_preventative_repair']) * self.g.costs.costlist[repairInfo['Repair']] #Find the difference between the cost of the full repair and the preventable repair
                        duration, vessel_cost, vessel_type = farm.OandM.vesselCost(repairInfo['Repair'])
                        corrective_cost += self.g.costs.costlist[repairInfo['Repair']] + vessel_cost
                    downtime = farm.OandM.downtime_todate()
                    print('\tAt {}, the total cost of O&M work to date is {:,} with a downtime of {:,} turbine-hours. There were {} trips to repair {} failures ({} repairs per trip.)'.format(farm.name, self.roundTo(farm.OandM.totalSpent,10000), self.roundTo(downtime,1000),months-1,len(farm.OandM.repairLog),round(len(farm.OandM.repairLog)/max(1,months),1)))
                    print('\t\tIf Corrective Maintenance would have been used to repair failures immediately, there would have been {} trips with a total cost of {:,} and 0 turbine-hours lost.'.format(len(farm.OandM.repairLog),self.roundTo(corrective_cost,10000)))
                    print('\t\tIf Condition Based Monitoring had been the strategy, {} full failures could have been prevented, possibly saving {:,}.\n'.format(preventablefails, self.roundTo(CBM_savings,10000)))

    def monthly_scheduled_estimator(self,farm,failures):
        repair_costs = 0
        monthly_vessel_cost = (0,)
        month_j = 0
        total_downtime=0
        for log in failures:
            t=log['TimeofFailure']
            month_of_failure=math.floor(t/43800)+1 #months in 1 thru 12
            if month_of_failure != month_j:
                repair_costs += self.g.costs.costlist[log['Repair']]
                try:
                    repair_costs += max(monthly_vessel_cost)
                except:
                    repair_costs += 0
                monthly_vessel_cost = (0,)
                duration, vessel_cost, vessel_type = farm.OandM.vesselCost(log['Repair'])
                monthly_vessel_cost += (vessel_cost,)
                month_j = month_of_failure
            elif month_of_failure == month_j:
                repair_costs += self.g.costs.costlist[log['Repair']]
                duration, vessel_cost, vessel_type = farm.OandM.vesselCost(log['Repair'])
                monthly_vessel_cost += (vessel_cost,)

            if 'Cell' in log['Repair Location']: #Do not account for failrues during construction, which are not repaired by monthly maintenance
                total_downtime += 0
            else:
                total_downtime += log['Number of turbines'] * ((month_of_failure)*43800-t)/60 #in turbine hours

        return repair_costs, int(total_downtime)

    def optimizeStructures(self,farm,turbinetype,structuretype_used): #check structure
        alternative_cost = {}
        alternative_failrates ={}
        baseline_failure_rates = mFailureRates.FailureRates()
        baseline_failure_rates.read()
        baseline_failure_rates.sort()

        ideal_structure_cost = structuretype_used
        ideal_structure_failure = structuretype_used

        ideal_cost = structuretype_used.priceQuote(self.g.costs.costlist, farm.myCell.depth, farm.myCell.soilqual)
        ideal_failurerate = sum(list(farm.mySubstructureList[0].myStructureFailureRates.values()))

        for structuretype in farm.substructureCatalog:
            cost = structuretype.priceQuote(self.g.costs.costlist, farm.myCell.depth, farm.myCell.soilqual)
            S = mSubstructure.Substructure(structuretype,1)
            S.setup(farm, baseline_failure_rates.structureFailureRates, turbinetype.cap, farm.myCell.soilqual,farm.myCell.meanH, farm.status['Scour Protection'])
            alternative_cost[structuretype] = cost
            alternative_failrates[structuretype] = S.myStructureFailureRates
            failurerate = sum(list(S.myStructureFailureRates.values()))

            if structuretype.type == 'gravity':
                if farm.myCell.depth > 30: #For gravity structures, dont consider reliable after 30m because of th eimpact of transportation. hard to capture otherwise
                    failurerate = 1
            if cost < ideal_cost:
                ideal_cost = cost
                ideal_structure_cost = structuretype
            if failurerate < ideal_failurerate:
                ideal_failurerate = failurerate
                ideal_structure_failure = structuretype

        if ideal_structure_cost != structuretype_used and ideal_structure_failure != structuretype_used:
            print('Design Tip: At {}, given the water depth, soil quality, wave heights, and turbine size, your choice of substructure ({}) was neither the most cost effective nor the most reliable. Check out the game instructions to see which substructures work best in these conditions.'.format(farm.name,structuretype_used.type))
        elif ideal_structure_cost == structuretype_used and ideal_structure_failure != structuretype_used:
            print('Design Tip: At {}, your choice of substructure ({}) was the most economical, but not the most reliable ({}).'.format(farm.name,structuretype_used.type,ideal_structure_failure.type))
        elif ideal_structure_cost != structuretype_used and ideal_structure_failure == structuretype_used:
            print('Design Tip: At {}, your choice of substructure ({}) was the most reliable, but not the most economical ({}).'.format(farm.name,structuretype_used.type,ideal_structure_cost.type))
        elif ideal_structure_cost == structuretype_used and ideal_structure_failure == structuretype_used:
            print('Design Tip: Nice job, at {} you chose the most economical and reliable substructure ({}) for this location.'.format(farm.name,structuretype_used.type))

        self.optimizedFarms.append(farm)

    def roundTo(self,value,nearest):
        return int(math.floor(value / nearest)) * nearest