""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license are defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""

from mTurbine import *
from mSubstructure import *
from mFailureRates import *
import math, numpy, mLCOE, timeit

class WindFarm:
    """Represents a wind farm"""

    def __init__(self,turbineList,structureList):
        self.name= None
        self.myCell = None
        self.myTurbineList=[]
        self.mySubstructureList=[]
        self.costs={'CAPEX':0,'OPEX':0} #Later transfer all costs here, to save space in code
        self.windSpeed=0
        self.waveHeight=1 #set at 1 to allow for repairs during construction, before simulations are run
        self.farmFailure={}
        self.failures={}
        self.notification = {}
        self.turbineCatalog=turbineList
        self.substructureCatalog=structureList
        self.capacityFactor=0.0
        self.installedCapacity=None
        self.cableLength=None #kilometers
        self.interarray_cableLength = 0  # meters
        self.connection=None
        self.myFailureRates=FailureRates() #Each farm has its own failure rates! can change based on actions of the game
        self.revenue=0
        self.energyOverLifetime = 0
        self.co2prevented_byfarm=0
        self.co2install=0
        self.breakEven={}
        self.contractorCost=0
        self.status={}
        self.constructiontime={}
        self.news={}
        self.lifetime = 20*525600 #years to hours
        self.OandM=None
        self.substation = None
        self.cables = None
        self.sellingPrice = {}
        self.LCOE = mLCOE.LCOE()
        ###########Some key figures for thesis findings###########
        self.availability = 1
        self.unavailableTime = {'farm':0,'turbine':0,'actual':0}
        self.lossEnergy = 0
        self.lossProduction = 0

    def design(self, **windFarmInput):
        self.myFailureRates.read()
        self.myFailureRates.sort()
        self.myCell=windFarmInput['cell']
        if windFarmInput['Additional_Costs']['Scour_protections'] == 0:
            self.status['Scour Protection']='None'
        else:
            self.status['Scour Protection'] = 'Rock Armouring'
        if windFarmInput['Additional_Costs']['Environmental Mitigation'] == 0:
            self.status['Environmentally Mitigated'] = False
        else:
            self.status['Environmentally Mitigated'] = True

        numTurbines = windFarmInput['numturbines']
        turbineIndex=windFarmInput['turbineIndex']
        structureIndex=windFarmInput['substructureIndex']
        self.myTurbineList=[]
        for i in range(numTurbines):
            turbinetype = self.turbineCatalog[turbineIndex-1]
            turbine=Turbine(turbinetype, i+1)
            turbine.setup(self.myFailureRates.turbineFailureRates)
            self.myTurbineList.append(turbine)     #This is where characteristics of wind farm are generated from input
        self.mySubstructureList=[]
        for j in range(numTurbines):
            structuretype=self.substructureCatalog[structureIndex - 1]
            structure=Substructure(structuretype, j + 1)

            structure.setup(self, self.myFailureRates.structureFailureRates, self.myTurbineList[0].type.cap, self.myCell.soilqual, self.myCell.meanH, self.status['Scour Protection'])
            self.mySubstructureList.append(structure)

        self.cableLength = round(math.sqrt(math.pow(self.myCell.xcoor,2)+math.pow(self.myCell.ycoor,2)),2) #In the game, this is checked for a shorter route by looking at neighboring farms
        self.substation = 'onshore' if self.cableLength <= 100 else 'offshore' #offshore substation required if more than 100km from shore
        self.cables = 'AC' if self.substation == 'onshore' else 'DC'
        self.interarray_cableLength = max(numTurbines*16/20,1.6065 * numTurbines - 16.065) * 1000 #meters #*= (numTurbines - 1)# * self.interarray_distance
        self.installedCapacity=numTurbines*(self.turbineCatalog[turbineIndex-1].cap)
        self.breakEven={'cost':False,'co2':0,'Profit':False}

        self.constructiontime['original'] = math.ceil((0.1*1440*self.cableLength +
                                                       numTurbines * 1440 * (0.65 + self.mySubstructureList[-1].type.installationDaysPerMW) +
                                                       numTurbines * 1440 * self.mySubstructureList[-1].type.fabricationDaysPer)/60) * 60   #0.65 days per turbine per MW math.ceil((2*1440*numTurbines+.2*1440*self.cableLength)/60)*60#in minutes: 2 days per turbine+cable length ()original
        self.constructiontime['actual'] = math.ceil((0.1*1440*self.cableLength +
                                                       numTurbines * 1440 * (0.65 + self.mySubstructureList[-1].type.installationDaysPerMW) +
                                                       numTurbines * 1440 * self.mySubstructureList[-1].type.fabricationDaysPer)/60) * 60   #0.65 days per turbine per MW math.ceil((2*1440*numTurbines+.2*1440*self.cableLength)/60)*60#in minutes: 2 days per turbine+cable length ()original

        self.status['life']='normal' #Used as an indicator during lifetime extension

        return {'cell': self.myCell, 'turbine': self.turbineCatalog[turbineIndex-1], 'numTurbines': int(numTurbines), 'cableLength':self.cableLength}

    def build(self,costperday,dt):
        self.newspaper(str(self.name) + ' under construction')
        self.costs['contractorCost']+=(dt*costperday/1440) #Cost per day for ten minutes
        if random.random() < self.myFailureRates.failureRates['Slow_contractor'] and self.farmFailure.get(self)==None:
            self.constructiontime['actual']+=1440 #Add one day to work!
            self.status['birthday']+=1440
            self.newspaper("You've got a slow contractor working on {}".format(self.name))
        Pf=self.myFailureRates.ConstFailureRates
        for item in Pf:
            if random.random() < Pf[item] and self.farmFailure=={}:
                self.notification['fail'] = 'new'
                self.farmFailure[self]=item
                self.newspaper('<!><!><!><!><!><!><!><!><!><!><!><!><!><!>\n <!>'+str(self.name) + ' has had a '+str(item)+' failure <!>\n<!> You will still be billed by the contractor for being on site. Better fix this soon!\n<!><!><!><!><!><!><!><!><!><!><!><!><!><!>\n')
                self.OandM.logFailures(self.myCell, item)
            elif self.farmFailure!={}: #If there is a construction failure/accident that hasn't been repaired yet, no work is being done
                self.constructiontime['actual'] += dt  # Add one day to work!
                self.status['birthday'] += dt

    def newspaper(self,news): #use one line function to publish news the first time it happens, if it keeps happening
        if news not in self.news:
            print(news)
            self.news[news] = 'oldnews'

    def checkFailure(self,int):            #For now, all farms have the same failure rates. this will be changed later on, as more dynamic risk elements are included
        Pf=self.myFailureRates.farmFailureRates
        if self.connection!=None:
            if ('farm_grid' in self.connection.farmFailure.values())==True or ('farm_grid_connection_failure' in self.connection.farmFailure.values())==True:  # Check if connecting wind farm has a grid failure (its own, or a third connectors)
                self.farmFailure[self.connection]='farm_grid_connection_failure' #Temporarily shut down this farm, but no need to repair seperately
                if 'chain_failure' not in self.connection.notification: #Check if this is a new failure#self.connection.notification['chain_failure']!='new' and self.connection.notification['chain_failure']!='old':
                    self.connection.notification['chain_failure']='new'
                    print("\nBecause {} has had a grid failure, {} is going offline\n".format(self.connection.name,self.name))
                    self.connection.notification['chain_failure']='old'
                self.unavailableTime['farm']+=int*len(self.myTurbineList) #For thesis report

            elif self.farmFailure.get(self.connection)!=None:#self.farmFailure[self.connection]=='farm_grid_connection_failure': #and self.connection.farmFailure[self.connection]!='farm_grid':
                del self.farmFailure[self.connection]
                print("Now that you've fixed the grid at {}, {} has come back online".format(self.connection.name,self.name))
        for item in Pf:
            if random.random() < Pf[item] and self.farmFailure.get(self)==None and self.farmFailure.get(self.connection)==None:
                self.farmFailure[self]=item         #Used to indicate that this farm has failed as a whole. need to record this as turbine will not be failing while whole farm is down.
                self.notification['fail']='new'     #Used so that the game may detect when a new failure has occured - and prompt the user to fix it immediately, or save for later
                self.failures[self]=item            #Add this whole farm failure to list of failures at this farm (may also include turbine failures)

    def run(self,clock,int):
        math.sqrt(self.status['age']) #If age is a negative number - wind farm hasn't been constructed yet, and run will move the exception, which is the building of the farm
        energy = 0
        self.checkFailure(int)
        month=math.ceil((clock/43800)%12) #get month ID (1 thru 12) from time in minutes
        if month==0: #At the moment the year turns, month is 0 not 1. quick fix.
            month=1

        start = timeit.default_timer()
        self.myCell.cellWeather.simulation(month,int)
        u90 = self.myCell.cellWeather.windSpeed_10min
        H = self.myCell.cellWeather.waveHeight
        self.windSpeed=round(numpy.mean(self.myCell.cellWeather.windSpeed_10min),2) #Mean hourly wind speed
        self.waveHeight=round(H,2)
        stop = timeit.default_timer()
        cf_hour = 0
        for structure in self.mySubstructureList:
            if structure.structureFailure== None: #Check if structural failure has occured
                structure.checkFailure()
                if structure.structureFailure!= None:               #If there is a structural failure
                    self.failures[structure] = structure.structureFailure # Add the structure failure to the list of failures at this farm (may also include failure of other structures and turbines and a failure of the farm as a whole. the failure of an entire farm does not stop structures from failing.
                    self.myTurbineList[structure.id-1].notification['autoshutdown']=True #auto shut down turbine whos structure has failed

        if self.farmFailure =={}: #Run turbines and check if failure has occured, if the farm is still up
            for turbine in self.myTurbineList:
                if turbine.turbineFailure==None and 'autoshutdown' not in turbine.notification:
                    turbine.checkFailure()#self.myFailureRates.turbineFailureRates)
                    if turbine.turbineFailure !=None:
                        self.failures[turbine]=turbine.turbineFailure    #Add the turbine failure to the list of failures at this farm (may also include failure of other turbines and a failure of the farm as a whole - after which the farm is shut off and no more turbine failures occur until repair.)

                elif 'autoshutdown' in turbine.notification or turbine.turbineFailure!=None:
                    self.unavailableTime['turbine'] += int  # for thesis reporting

                for u in u90:#Six wind speeds sampled per hour
                    power = turbine.run(u)
                    energy += power*(1/len(u90)) #power times time

            cf_hour=energy/(self.installedCapacity*int/60) #MW to MWhr for whole farm
        elif self.farmFailure !={} and self.unavailableTime['farm']==0:
            self.unavailableTime['farm'] += int * len(self.myTurbineList) #For thesis report


        ##########Find availability##########
        report_energy = 0
        for u in u90:
            power = self.myTurbineList[0].productionLoss(u)
            report_energy += power * (1 / len(u90))
        if self.unavailableTime['farm'] ==0: #Avoid double counting when there are failures and when farm is down
            self.unavailableTime['actual'] += self.unavailableTime['turbine']  #accounts for structural failures as well since turbines are "autoshut down" after a structural failure
            N=self.unavailableTime['turbine']/int
            self.lossEnergy+=report_energy*N
        else:
            self.unavailableTime['actual'] += self.unavailableTime['farm']
            self.lossEnergy+=report_energy*len(self.myTurbineList)

        self.lossProduction = round(self.lossEnergy*self.sellingPrice['First10Years']/(self.status['age']/525600),4)
        self.unavailableTime['farm'] = 0
        self.unavailableTime['turbine'] = 0
        self.availability = round(1 - self.unavailableTime['actual']/(self.status['age'] * len(self.myTurbineList)),4)
        #####################################

        iteration = self.status['age'] / int
        self.capacityFactor = round((self.capacityFactor * (iteration -1) + cf_hour)/iteration,1)
        self.energyOverLifetime += energy
        return {'energy':energy,'capacity factor':self.capacityFactor,'windSpeed':self.windSpeed,'waveHeight':self.waveHeight,'simulation time':stop-start}

    def updateLCOE(self):
        self.LCOE.update(self.costs['CAPEX'],self.costs['OPEX'],self.energyOverLifetime)
        # print(self.costs['CAPEX'])
        # print(self.costs['OPEX'])
        # print(self.energyOverLifetime)

    def display(self):
        bdayyear = math.floor(self.status['birthday']/(60*24*365)) + 2018
        bdaymonth = math.floor(((self.status['birthday'] / (60 * 24 * 365)) - (bdayyear-2018)) * 12) +1

        ageyears = round(self.status['age']/(60*24*365),1)
        lifetimeyears = round(self.lifetime/(60*24*365),1)
        print('\t' + str(self.name) + ' : Cell ' + str(self.myCell.id) +' : Cost €' + str(self.roundTo(self.costs['CAPEX'],100000)/1000000) +
              ' mil : Turbine ' + str(self.myTurbineList[0].type.cap) + 'MW on ' + str(self.mySubstructureList[-1].type.type) +
              ' : No. turbines ' + str(len(self.myTurbineList)) + ' : Cable Length ' + str(int(self.cableLength)) +'km : O&M strategy ' +
              str(self.OandM.strategy) +': Capacity Factor ' + str(self.capacityFactor) +' : Revenue €' +
              str(self.roundTo(self.revenue, 100000)/1000000) + ' mil : Status ' + str(self.status['phase']) +
              '\tBirthday ' + str(bdaymonth) + '-' + str(bdayyear)+' : Age ' + str(ageyears) + 'yrs : Lifetime ' + str(lifetimeyears) + 'yrs : Availability [%] ' + str(self.availability)+
              ' : Production losses annual: €' + str(self.lossProduction) + ' : Annual O&M Cost: €' + str(self.OandM.annualOandMcost),' : LCOE €/MWh' + str(self.LCOE.value))

    def roundTo(self,value,nearest):
        return int(math.ceil(value / nearest)) * nearest

